import bpf_asm


TEST_BPF = [(40, 0, 0, 12), (21, 0, 1, 2048), (6, 0, 0, 65536), (6, 0, 0, 0)]
TEST_BPF_ASM = """\
    ldh [12]
    jeq #0x800, accept, drop
accept:
    ret #65536
drop:
    ret #0
"""

def test_assemble():
    assert bpf_asm.assemble(TEST_BPF_ASM) == TEST_BPF


def test_dumps_default():
    assert bpf_asm.dumps(TEST_BPF) == "4,40 0 0 12,21 0 1 2048,6 0 0 65536,6 0 0 0"


def test_dumps_multiline():
    assert bpf_asm.dumps(TEST_BPF, bpf_asm.BpfDumpType.MULTILINE) == "4\n40 0 0 12\n21 0 1 2048\n6 0 0 65536\n6 0 0 0"


def test_dumps_c_array():
    assert bpf_asm.dumps(TEST_BPF, bpf_asm.BpfDumpType.C_ARRAY) == \
        "{ 0x28, 0, 0, 0x0000000c },\n{ 0x15, 0, 1, 0x00000800 },\n{ 0x6, 0, 0, 0x00010000 },\n{ 0x6, 0, 0, 0x00000000 },"



"""
Compile BPF assembly.
"""
import sys
import argparse

import bpf_asm


__package__ = "bpf_asm"


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    format_group = parser.add_mutually_exclusive_group()
    format_group.add_argument('-f', '--format', default='DEFAULT',
                              choices=list(i.name for i in bpf_asm.BpfDumpType),
                              help="format to dump in")
    format_group.add_argument('-c', action='store_const', dest='format',
                              const='C_ARRAY', help='print C style')
    format_group.add_argument('-p', '--python', action='store_true',
                              help="dump as a Python list")
    parser.add_argument('input', type=argparse.FileType('r'),
                        help="bpf to assemble ('-' for stdin)")

    args = parser.parse_args()

    bpf = bpf_asm.assemble(args.input.read())

    if args.python:
        print(repr(bpf))
    else:
        print(bpf_asm.dumps(bpf, bpf_asm.BpfDumpType[args.format]))


if __name__ == "__main__":
    sys.exit(main())



# pybpf_asm - Python BPF Assembler
# Copyright (C) 2022  Segev Finer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# cython: language_level=3str, binding=True
"""
Python BPF Assembler.

Based on the BPF assembler in Linux sources.
"""

import enum
from typing import List, Tuple
from libc.stdlib cimport free
from libc.stdint cimport uint8_t, uint16_t, uint32_t


class Error(Exception):
    """Raised when there is an error assembling BPF."""
    pass


class BpfDumpType(enum.Enum):
    DEFAULT = 0
    MULTILINE = 1
    C_ARRAY = 2


cdef extern from "bpf_exp.yacc.h" nogil:
    struct sock_filter:
        uint16_t	code   # Actual filter code
        uint8_t	    jt	   # Jump true
        uint8_t	    jf	   # Jump false
        uint32_t	k      # Generic multiuse field

    int bpf_asm_compile(const char *str, int len, sock_filter (**out), const char **error)


def assemble(str: str) -> List[Tuple[int, int, int, int]]:
    """Assemble BPF *str*."""
    cdef sock_filter *out = NULL
    cdef char *error = NULL

    strbytes = str.encode()
    strbytes_ptr = <const char*>strbytes
    strbytes_len = len(strbytes)
    with nogil:
        result = bpf_asm_compile(strbytes_ptr, strbytes_len, &out, &error)
    try:
        if not result:
            raise Error(error.decode())

        insns = []
        for insn in out[:result]:
            insns.append((insn.code, insn.jt, insn.jf, insn.k))
        return insns
    finally:
        free(out)
        free(error)


def dumps(bpf: List[Tuple[int, int, int, int]], format=BpfDumpType.DEFAULT) -> str:
    """Dump *bpf* to a str in *format*."""
    result = []

    if format == BpfDumpType.DEFAULT:
        result.append(f"{len(bpf)},")

        for insn in bpf[:-1]:
            result.append(f"{insn[0]} {insn[1]} {insn[2]} {insn[3]},")

        insn = bpf[-1]
        result.append(f"{insn[0]} {insn[1]} {insn[2]} {insn[3]}")

        return ''.join(result)

    elif format == BpfDumpType.MULTILINE:
        result.append(f"{len(bpf)}")
        for insn in bpf:
            result.append(f"{insn[0]} {insn[1]} {insn[2]} {insn[3]}")
        return '\n'.join(result)

    elif format == BpfDumpType.C_ARRAY:
        for insn in bpf:
            result.append(f"{{ 0x{insn[0]:x}, {insn[1]}, {insn[2]}, 0x{insn[3]:08x} }},")
        return '\n'.join(result)

    else:
        raise ValueError(f"unknown format: {format!r}")



# pybpf_asm - Python BPF Assembler
# Copyright (C) 2022  Segev Finer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Python BPF Assembler.

Based on the BPF assembler in Linux sources.
"""

from ._bpf_asm import *


__version__ = "0.1.1"



# pybpf_asm - Python BPF Assembler
# Copyright (C) 2022  Segev Finer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import re
from setuptools import find_packages
from skbuild import setup


with open("bpf_asm/__init__.py", "r", encoding="utf-8") as f:
    version = re.search(r'(?m)^__version__ = "([a-zA-Z0-9.-]+)"', f.read()).group(1)

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="bpf_asm",
    version=version,
    author="Segev Finer",
    author_email="segev208@gmail.com",
    description="Python BPF Assembler",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/segevfiner/pybpf_asm",
    project_urls={
        "Documentation": "https://segevfiner.github.io/pybpf_asm/",
        "Issue Tracker": "https://github.com/segevfiner/pybpf_asm/issues",
    },
    license="GPL-2.0-only",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: System :: Networking :: Monitoring",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="bpf",
    zip_safe=False,
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "pybpf_asm = bpf_asm.__main__:main"
        ],
    },
    python_requires='>=3.6',
    extras_require={
        "dev": [
            "flake8",
            "pytest",
            "sphinx==4.*"
        ],
    }
)
